!$Id: b.f90,v 1.1.1.1 2004/02/10 19:12:06 steve Exp $
#include "cctk.h"

!---------------------------------------------------------------------
!
!     SUBROUTINE CALB
!
!     This subroutine defines the vector "b" in terms of the shift.
!     This subroutine receives:
!        (1) The shift (index up),
!        (2) Spatial derivatives of the contravariant shift,
!        (3) The position,
!        (4) The parameter array.
!
!     This subroutine needs to define:
!        (1) The vector b (index up)
!        (2) Spatial derivatives of contravariant b
!
!     To switch between different choices of b use the "btype" variable
!        btype = 1:  Daves set of equations with pi = diff(phi,t)
!        btype = 2:  Oliviers first set of equations with 
!                       pi = diff(phi,t) - beta^m*diff(phi,x^m)
!        btype = 3:  b = f*beta, where f is a radial function.
!
!---------------------------------------------------------------------
      subroutine calB(beta_up,diff_beta_up,b_up,diff_b_up,xx,par)
      use params
      implicit   none

      CCTK_REAL, dimension(3), intent(in) :: beta_up
      CCTK_REAL, dimension(3,3), intent(in) :: diff_beta_up
      CCTK_REAL, dimension(3), intent(out) :: b_up
      CCTK_REAL, dimension(3,3), intent(out) :: diff_b_up
      CCTK_REAL, intent(in) :: xx(3), par(NPAR)

      !   local vars
      CCTK_REAL  r, x, y, z, rbmax, rbmin
      CCTK_REAL ::  xmax,f
      CCTK_INT  :: i,j, btype
      CCTK_REAL, dimension(3) :: diff_f
      CCTK_REAL :: a0, a1, a2, a3, a4, a5, drf, x1, x2

      btype = int(par(P_SHIFT_TYPE))
      rbmin = par(P_RBMIN)
      rbmax = par(P_RBMAX)

      x = xx(1)
      y = xx(2)
      z = xx(3)

      if (btype .eq. 1) then
!        --------------------------------------------
!        b = 0 (This is Daves set of equations)
!        --------------------------------------------
         b_up = 0.0
         diff_b_up = 0.0

      else if (btype .eq. 2) then
!        --------------------------------------------
!        b = beta (This is Oliviers first set of eqs.
!        --------------------------------------------
         b_up = beta_up
         diff_b_up = diff_beta_up

      else if (btype .eq. 3) then
!        --------------------------------------------
!        b is product of beta and a radial function.
!            b^i = f(r)*beta^i
!        --------------------------------------------
         r = sqrt(x**2+y**2+z**2)
         
         if (r.lt.rbmin) then
            f = 1.0
            diff_f = 0.0
         else if (r.le.rbmax.and.r.ge.rbmin) then

            ! The interpolation function is
            !   f = a0 + a1*x + a2*x^2 + a3*x^3 + a4*x^4 + a5*x^5
            !
            ! The conditions on f are that
            ! f(x1) = 1
            ! f(x2) = 0
            ! df/dx = 0 at x1 and x2
            ! d^2f/dx^2 = 0 at x1 and x2

            x2 = rbmax
            x1 = rbmin
            a0 = -x2**3*(x2**2+10.d0*x1**2-5.d0*x2*x1)/(x1**5-5.d0*x2*x1**4&
                 +10.d0*x2**2*x1**3-x2**5-10.d0*x1**2*x2**3+5.d0*x2**4*x1)
            a1 = 30.d0*x2**2*x1**2/(x1**5-5.d0*x2*x1**4+10.d0*x2**2*x1**3&
                 -x2**5-10.d0*x1**2*x2**3+5.d0*x2**4*x1)
            a2 = -30.d0*x1*x2*(x1+x2)/(x1**5-5.d0*x2*x1**4&
                 +10.d0*x2**2*x1**3-x2**5-10.d0*x1**2*x2**3+5.d0*x2**4*x1)
            a3 = 10.d0*(x1**2+4.d0*x2*x1+x2**2)/(x1**5-5.d0*x2*x1**4&
                 +10.d0*x2**2*x1**3-x2**5-10.d0*x1**2*x2**3+5.d0*x2**4*x1)
            a4 = -15.d0*(x1+x2)/(x1**5-5.d0*x2*x1**4+10.d0*x2**2*x1**3&
                 -x2**5-10.d0*x1**2*x2**3+5.d0*x2**4*x1)
            a5 = 6.d0/(x1**5-5.d0*x2*x1**4+10.d0*x2**2*x1**3-x2**5&
                 -10.d0*x1**2*x2**3+5.d0*x2**4*x1)
            f = a0 + r*(a1 + r*(a2 + r*(a3 + r*(a4 + a5*r))))
            drf = a1 + r*(2.d0*a2 + r*(3.d0*a3 + r*(4.d0*a4 + 5.*a5*r)))
            diff_f(1) = drf*x/r
            diff_f(2) = drf*y/r
            diff_f(3) = drf*z/r

         else
            f = 0.0
            diff_f = 0.0
         end if
         do i=1,3
            b_up(i) = f*beta_up(i)
         end do
         do i=1,3
            do j=1,3
               diff_b_up(i,j) = f*diff_beta_up(i,j) + beta_up(i)*diff_f(j)
            end do
         end do
         
      else
         write(0,*)'Unknown btype ',btype
         stop
      end if
   
      return
      end

