
# include "cctk.h"

!$Id: metric.f90,v 1.1.1.1 2004/02/10 19:12:06 steve Exp $

module background_quant
use params
implicit none

contains
!--------------------------------------------------------------------------
!
! Compute background quantities that are needed in rhs, and to define the 
! extra vars that need discrete derivatives. Notice that, unlike discrete derivatives,
! here for analytic derivatives we use diff_bla(...,j), where j, the last index,
! is the direction in which the derivative is taken. The quantities here computed:
!
!     sdet = sqrt of the determinant of the 3-metric (indexes down)
!     alpha = lapse
!     beta_up(i) = shift with index up
!     b_up(i) = "b" interpolating function, index up
!     h_up(i,j) = 3-metric, indexes up
!     g_up(i,j) = 4-metric, indexes up
!     h(i,j) = 3-metric, indexes down
!     g(i,j) = 4-metric, indexes down
!     aux1(j) = beta^i*partial_ib^j-b^i*partial_ibeta^j
!     diff_aux2(j,i) = analytic derivative in the "i" direction of 
!                      aux2(j) = sdet*b^j/alpha
!     diff_aux3(j,i) = analytic derivative in the "i" direction of
!                      aux3(i) = sdet*b^i

subroutine background(xx,sdet,alpha,beta_up,b_up,aux1,h_up,g_up,h,g,diff_aux2,&
     diff_aux3,diff_b_up,par)

  implicit none

  CCTK_REAL, dimension(3), intent(in):: xx
  CCTK_REAL, intent(out) :: alpha,sdet ! scalars
  CCTK_REAL, dimension(3), intent(out) :: beta_up,b_up,aux1 ! vectors
  CCTK_REAL, dimension(3,3),intent(out):: h_up,h,diff_aux2,diff_aux3 ! tensors
  CCTK_REAL, dimension(4,4),intent(out):: g,g_up ! more tensors
  CCTK_REAL, dimension(NPAR),intent(in)  :: par
  ! local vars
  CCTK_REAL :: r,x,y,z,m
  CCTK_INT :: i,j
  CCTK_REAL, dimension(3,3) :: diff_beta_up, diff_b_up
  CCTK_REAL, dimension(3) :: diff_sdet, diff_alpha

  m = par(P_BH1_MASS)

  if (m .ne. 0.0) then
  x = xx(1) - par(P_BH1_X0)
  y = xx(2) - par(P_BH1_Y0)
  z = xx(3) - par(P_BH1_Z0)

  r=sqrt(x**2+y**2+z**2)
  alpha = 1./sqrt(1.+2.*m/r)

  beta_up(1) = 2.*m/r*alpha**2*x/r
  beta_up(2) = 2.*m/r*alpha**2*y/r
  beta_up(3) = 2.*m/r*alpha**2*z/r

  h_up(1,1) = 1.-alpha**2*2.*m/r*x/r*x/r
  h_up(1,2) = -alpha**2*2.*m/r*x/r*y/r
  h_up(2,1) = h_up(1,2)
  h_up(1,3) = -alpha**2*2.*m/r*x/r*z/r
  h_up(3,1) = h_up(1,3)
  h_up(2,2) = 1.-alpha**2*2.*m/r*y/r*y/r
  h_up(2,3) = -alpha**2*2.*m/r*y/r*z/r
  h_up(3,2) = h_up(2,3)
  h_up(3,3) = 1.-alpha**2*2.*m/r*z/r*z/r

  ! the 3-metric, indexes down
  h(1,1) = 1.+2.*m/r*x/r*x/r
  h(1,2) = 2.*m/r*x/r*y/r
  h(2,1) = h(1,2)
  h(1,3) = 2.*m/r*x/r*z/r
  h(3,1) = h(1,3)
  h(2,2) = 1.+2.*m/r*y/r*y/r
  h(2,3) = 2.*m/r*y/r*z/r
  h(3,2) = h(2,3)
  h(3,3) = 1.+2.*m/r*z/r*z/r

  ! the 4-metric, indexes down
  do i=1,3
     do j=1,3
        g(i,j) = h(i,j)
     end do
  end do
  g(4,4) = -1.+2.*m/r
  g(1,4) = 2.*m/r*x/r
  g(4,1) = g(1,4)
  g(2,4) = 2.*m/r*y/r
  g(4,2) = g(2,4)
  g(3,4) = 2.*m/r*z/r
  g(4,3) = g(3,4)

  ! the 4-metric, indexes up, computed with maple in gamma_kerr_schild.map
  g_up(1,1) = (r**3-2*r**2*m+2*m*z**2+2*m*y**2)/(r**3-2*r**2*m+2*m*z**2+2*m*y**2+2*m*x**2)
  g_up(1,2) = -2*m*x*y/(r**3-2*r**2*m+2*m*z**2+2*m*y**2+2*m*x**2)
  g_up(1,3) = -2*m*x*z/(r**3-2*r**2*m+2*m*z**2+2*m*y**2+2*m*x**2)
  g_up(1,4) = 2*m*r*x/(r**3-2*r**2*m+2*m*z**2+2*m*y**2+2*m*x**2)
  g_up(2,1) = g_up(1,2)
  g_up(2,2) = (r**3-2*r**2*m+2*m*z**2+2*m*x**2)/(r**3-2*r**2*m+2*m*z**2+2*m*y**2+2*m*x**2)
  g_up(2,3) = -2*m*y*z/(r**3-2*r**2*m+2*m*z**2+2*m*y**2+2*m*x**2)
  g_up(2,4) = 2*m*r*y/(r**3-2*r**2*m+2*m*z**2+2*m*y**2+2*m*x**2)
  g_up(3,1) = g_up(1,3)
  g_up(3,2) = g_up(2,3)
  g_up(3,3) = (r**3-2*r**2*m+2*m*y**2+2*m*x**2)/(r**3-2*r**2*m+2*m*z**2+2*m*y**2+2*m*x**2)
  g_up(3,4) = 2*m*r*z/(r**3-2*r**2*m+2*m*z**2+2*m*y**2+2*m*x**2)
  g_up(4,1) = g_up(1,4)
  g_up(4,2) = g_up(2,4)
  g_up(4,3) = g_up(3,4)
  g_up(4,4) = -(r**3+2*m*z**2+2*m*y**2+2*m*x**2)/(r**3-2*r**2*m+2*m*z**2+2*m*y**2+2*m*x**2)


  ! sqrt of the determinant of the 3-metric (indexes down)

  sdet = sqrt(h(1,1)*(h(2,2)*h(3,3)-h(2,3)*h(3,2)) - h(2,1)*(h(1,2)*h(3,3)-h(1,3)*h(3,2)) + &
       h(3,1)*(h(1,2)*h(2,3)-h(1,3)*h(2,2)))

  
  !-------- analytically compute some background quantities -----------!
  ! derivatives of sdet, done with maple in shift.map
  diff_sdet(1) = 1/sqrt(1/sqrt(x**2+y**2+z**2)**3*(2*m*x**2+2*m*y**2&
       +2*m*z**2+sqrt(x**2+y**2+z**2)**3))*(-3/sqrt(x**2+y**2+z**2)**5*(2&
       *m*x**2+2*m*y**2+2*m*z**2+sqrt(x**2+y**2+z**2)**3)*x+1/sqrt(x**2+y&
       **2+z**2)**3*(4*x*m+3*sqrt(x**2+y**2+z**2)*x))/2
  
  diff_sdet(2) = 1/sqrt(1/sqrt(x**2+y**2+z**2)**3*(2*m*x**2+2*m*y**2&
       +2*m*z**2+sqrt(x**2+y**2+z**2)**3))*(-3/sqrt(x**2+y**2+z**2)**5*(2&
       *m*x**2+2*m*y**2+2*m*z**2+sqrt(x**2+y**2+z**2)**3)*y+1/sqrt(x**2+y&
       **2+z**2)**3*(4*m*y+3*sqrt(x**2+y**2+z**2)*y))/2
  
  diff_sdet(3) = 1/sqrt(1/sqrt(x**2+y**2+z**2)**3*(2*m*x**2+2*m*y**2&
       +2*m*z**2+sqrt(x**2+y**2+z**2)**3))*(-3/sqrt(x**2+y**2+z**2)**5*(2&
       *m*x**2+2*m*y**2+2*m*z**2+sqrt(x**2+y**2+z**2)**3)*z+1/sqrt(x**2+y&
       **2+z**2)**3*(4*m*z+3*sqrt(x**2+y**2+z**2)*z))/2
  ! derivatives of alpha, done with maple in shift.map
  diff_alpha(1) = 1/sqrt((sqrt(x**2+y**2+z**2)+2*m)/sqrt(x**2+y**2+z&
       **2))**3*m/sqrt(x**2+y**2+z**2)**3*x

  diff_alpha(2) = 1/sqrt((sqrt(x**2+y**2+z**2)+2*m)/sqrt(x**2+y**2+z&
       **2))**3*m/sqrt(x**2+y**2+z**2)**3*y

  diff_alpha(3) = 1/sqrt((sqrt(x**2+y**2+z**2)+2*m)/sqrt(x**2+y**2+z&
       **2))**3*m/sqrt(x**2+y**2+z**2)**3*z
  
  ! Derivatives of the shift, done with maple in shift.map (second index indicates the 
  ! direction of differentiation)
  
  diff_beta_up(1,1) = -4*m/(x**2+y**2+z**2)**2/(1+2*m/sqrt(x**2+y**2+z**2))*&
       x**2+4*m**2/sqrt(x**2+y**2+z**2)**5/(1+2*m/sqrt(x**2+y**2+z**2))**&
       2*x**2+2*m/(x**2+y**2+z**2)/(1+2*m/sqrt(x**2+y**2+z**2))
  
  diff_beta_up(2,1) = -4*m/(x**2+y**2+z**2)**2/(1+2*m/sqrt(x**2+y**2+z**2))*&
       x*y+4*m**2/sqrt(x**2+y**2+z**2)**5/(1+2*m/sqrt(x**2+y**2+z**2))**2*x*y
  
  diff_beta_up(3,1) = -4*m/(x**2+y**2+z**2)**2/(1+2*m/sqrt(x**2+y**2+z**2))*&
       x*z+4*m**2/sqrt(x**2+y**2+z**2)**5/(1+2*m/sqrt(x**2+y**2+z**2))**2*x*z
  
  diff_beta_up(1,2) = -4*m/(x**2+y**2+z**2)**2/(1+2*m/sqrt(x**2+y**2+z**2))*&
       x*y+4*m**2/sqrt(x**2+y**2+z**2)**5/(1+2*m/sqrt(x**2+y**2+z**2))**2*x*y
  
  diff_beta_up(2,2) = -4*m/(x**2+y**2+z**2)**2/(1+2*m/sqrt(x**2+y**2+z**2))*&
       y**2+4*m**2/sqrt(x**2+y**2+z**2)**5/(1+2*m/sqrt(x**2+y**2+z**2))**&
       2*y**2+2*m/(x**2+y**2+z**2)/(1+2*m/sqrt(x**2+y**2+z**2))

  diff_beta_up(3,2) = -4*m/(x**2+y**2+z**2)**2/(1+2*m/sqrt(x**2+y**2+z**2))*&
       y*z+4*m**2/sqrt(x**2+y**2+z**2)**5/(1+2*m/sqrt(x**2+y**2+z**2))**2*y*z
  
  diff_beta_up(1,3) = -4*m/(x**2+y**2+z**2)**2/(1+2*m/sqrt(x**2+y**2+z**2))*&
       x*z+4*m**2/sqrt(x**2+y**2+z**2)**5/(1+2*m/sqrt(x**2+y**2+z**2))**2*x*z

  diff_beta_up(2,3) = -4*m/(x**2+y**2+z**2)**2/(1+2*m/sqrt(x**2+y**2+z**2))*&
       y*z+4*m**2/sqrt(x**2+y**2+z**2)**5/(1+2*m/sqrt(x**2+y**2+z**2))**2*y*z
  
  diff_beta_up(3,3) = -4*m/(x**2+y**2+z**2)**2/(1+2*m/sqrt(x**2+y**2+z**2))*&
       z**2+4*m**2/sqrt(x**2+y**2+z**2)**5/(1+2*m/sqrt(x**2+y**2+z**2))**&
       2*z**2+2*m/(x**2+y**2+z**2)/(1+2*m/sqrt(x**2+y**2+z**2))
  
  ! compute b
  call calB(beta_up,diff_beta_up,b_up,diff_b_up,xx,par)
  
  ! compute aux1(j) = beta^i*partial_ib^j-b^i*partial_ibeta^j
  aux1=0.
  do j=1,3
     do i=1,3
        aux1(j) = aux1(j) + beta_up(i)*diff_b_up(j,i)-b_up(i)*diff_beta_up(j,i)
     end do
  end do
  
    ! derivatives of  aux2(i) = sdet*b^i/alpha
  do i=1,3
     do j=1,3
        diff_aux2(i,j) = diff_sdet(j)*b_up(i)/alpha + sdet*diff_b_up(i,j)/alpha - &
             sdet*b_up(i)/alpha**2*diff_alpha(j)
     end do
  end do
  
  ! compute derivatives of aux3(i) = sdet*b^i
  do i=1,3
     do j=1,3
        diff_aux3(i,j) = diff_sdet(j)*b_up(i) + sdet*diff_b_up(i,j)
     end do
  end do

  else
    r=sqrt(x**2+y**2+z**2)
    alpha = 1.0d0

    do i = 1, 3
      beta_up(i)    = 0.0d0
      b_up(i)       = 0.0d0
      diff_sdet(i)  = 0.0d0
      diff_alpha(i) = 0.0d0
      aux1(i)       = 0.0d0
    end do

    do i = 1, 3
    do j = 1, 3
      h_up(i,j)         = 0.0d0
      h(i,j)            = 0.0d0
      diff_beta_up(i,j) = 0.0d0
      diff_b_up(i,j)    = 0.0d0
      diff_aux2(i,j)    = 0.0d0
      diff_aux3(i,j)    = 0.0d0
    end do
    end do

    do i = 1, 4
    do j = 1, 4
      g(i,j)    = 0.0d0
      g_up(i,j) = 0.0d0
    end do
    end do

    do i = 1, 3
      h(i,i) = 1.0d0
      h_up(i,i) = 1.0d0
    end do

    do i = 1, 4
      g(i,i) = 1.0d0
      g_up(i,i) = 1.0d0
    end do

    sdet = 1.0d0

  end if

end subroutine background

end module background_quant
